'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
' Game15 v2, a simple "15 tiles" brick game, written in PB/DLL 6.0
' January 2000 by Borje Hagsten. Released as Public Domain.
' Updated June 2003 with better mix code - now shuffles bricks instead,
' plus added keyboard handling - can be played using the arrow keys.
' Free to enhance, learn from, or simply to enjoy and play.
'
' The "rules" are simple. Click on the tiles next to the open
' space in order to move them and try to line them up, 1 to 15.
' A beep and red caption on a brick indicates a correct position.
' It's not as easy as it may look like from the beginning and some
' people may even become addicted to playing it. A complete 32-bit
' game for Windows with a file size of just 20 KB - PB rules!
'
' - Some ideas, in case you want to do something more of it:
' Use a split bitmap instead of numbers on the labels
' Create and add resource file with version info and program icon
' Create routines to save and show results (maybe even store/recall a running game?)
' Add a "About" dialog, menu, sound support, help and some game info
' Set a price and sell a million copies (but please remember me, if you succeed..
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#Compile Exe
#Include "WIN32API.INC"

%ID_MIX    = 11
%ID_LABEL1 = 20
%ID_LABEL2 = 21
%ID_LABEL3 = 22
%ID_TIMER  = 40

Type bx
  x As Long ' x pos
  y As Long ' y pos
  z As Long ' reference number
  p As Long ' actual position
  c As Long ' color value
End Type

Global hDlg     As Long
Global gTime    As Long
Global brc      As Long
Global n        As Long
Global EmptyPos As Long
Global hTimer   As Long
Global bPos()   As bx

Declare CallBack Function DlgCallback()
Declare Function CheckDone As Long
Declare Function EndProcedure As Long
Declare Sub NewGame
Declare Sub ReMix

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
' PBMAIN, where all controls and the dialog is created
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Function PBMain
  Local x As Long, y As Long, z As Long, w As Long, h As Long, b As Long
  Local a As Long
  brc = 4        'number of columns/rows - use at least 3, classic is 4 (15-tile game)
' Note: The brc value is all you have to change. The dialog and all
' controls will resize themselves accordingly. Try with 5 or 6 to
' create a game with 25 or 36 bricks, but do have in mind that the
' time it takes to solve the puzzle relates to this number..  :-)

  w = 26          'width
  h = 24          'height
  a = 10          'distance from dialog edge
  n = brc^2 - 1   'number of bricks (tiles, whatever..)
  b = brc - 1     'zero based number of columns and rows
  EmptyPos = n    'initial empty space, down in the corner
  ReDim bPos (n)  'array for storing brick info

  Dialog New 0, Format$(n) & " tiles",,, brc * w + 22, brc * h + 90, _
                             %WS_CAPTION Or %WS_MINIMIZEBOX Or %WS_SYSMENU, 0 To hDlg

  Control Add Label,  hDlg, %ID_LABEL2, "",       9, 9, _
                      brc * w + 2, brc * h + 2, %SS_SUNKEN Or %WS_TABSTOP
  Control Add Label,  hDlg, %ID_LABEL1, "",       a, brc * h + 18, _
                      brc * w + 2, 12, %SS_CENTER  Or %SS_SUNKEN
  Control Add Label,  hDlg, %ID_LABEL3, "",       a, brc * h + 30, _
                      brc * w + 2, 12, %SS_CENTER  Or %SS_SUNKEN
  Control Add Button, hDlg, %ID_MIX,   "&New game", a, brc * h + 50, brc * w + 2, 14
  Control Add Button, hDlg, %IDCANCEL, "E&xit",     a, brc * h + 68, brc * w + 2, 14

  'A zero based index would be better here, but some low values are already
  'used by the DDT dialog (like IDCANCEL, etc), so we better use another
  'approach and increase it to a "100 based" line of controls instead
  z = 100
  For y = 0 To b
    For x = 0 To b
      bPos(z - 100).x = a + x * w
      bPos(z - 100).y = a + y * h
      Control Add Label, hDlg, z, Format$(z - 99), a + x * w, a + y * h, w, h, _
                         %SS_NOTIFY Or %SS_CENTER, %WS_EX_DLGMODALFRAME
      Incr z
      If z > 99 + n Then Exit For
    Next
  Next
  bPos(n).x = a + b * w  'x and y for the inital empty space
  bPos(n).y = a + b * h

  Dialog Show Modeless hDlg, Call DlgCallback

  'DDT dialog, so use own message loop to trap WM_KEYDOWN..
  Local msg As TAGMSG

  While GetMessage(Msg, %NULL, 0, 0)
     If IsDialogMessage(hDlg, Msg) Then
        Select Case msg.message
           Case %WM_CREATE
           Case %WM_CHAR : MessageBeep &HFFFFFFFF
           Case %WM_KEYDOWN
              If GetDlgCtrlID(msg.hWnd) <> %ID_LABEL2 Then Exit Select
              Select Case msg.wParam
                 Case %VK_LEFT
                    If (EmptyPos + 1) Mod brc <> 0 Then
                       For x = 0 To n - 1
                          If bPos(bPos(x).p).z = EmptyPos + 1 Then
                             x = 100 + bPos(x).p
                             SendMessage hDlg, %WM_COMMAND, MakLng(x, %BN_CLICKED), GetDlgItem(hDlg, x)
                             Exit For
                          End If
                       Next
                    End If

                 Case %VK_RIGHT
                    If EmptyPos Mod brc <> 0 Then
                       For x = 0 To n - 1
                          If bPos(bPos(x).p).z = EmptyPos - 1 Then
                             x = 100 + bPos(x).p
                             SendMessage hDlg, %WM_COMMAND, MakLng(x, %BN_CLICKED), GetDlgItem(hDlg, x)
                             Exit For
                          End If
                       Next
                    End If

                 Case %VK_UP
                    If EmptyPos + brc <= n Then
                       For x = 0 To n - 1
                          If bPos(bPos(x).p).z = EmptyPos + brc Then
                             x = 100 + bPos(x).p
                             SendMessage hDlg, %WM_COMMAND, MakLng(x, %BN_CLICKED), GetDlgItem(hDlg, x)
                             Exit For
                          End If
                       Next
                    End If

                 Case %VK_DOWN
                    If EmptyPos - brc >= 0 Then
                       For x = 0 To n - 1
                          If bPos(bPos(x).p).z = EmptyPos - brc Then
                             x = 100 + bPos(x).p
                             SendMessage hDlg, %WM_COMMAND, MakLng(x, %BN_CLICKED), GetDlgItem(hDlg, x)
                             Exit For
                          End If
                       Next
                    End If
              End Select
        End Select
     Else
        TranslateMessage Msg
        DispatchMessage  Msg
     End If
  Wend

End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
' One dialog Callback function takes care of almost everything
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
CallBack Function DlgCallback()
  Local  rStr         As String
  Static hBrushDialog As Long

  Select Case CbMsg
     Case %WM_INITDIALOG
        hBrushDialog = CreateSolidBrush(RGB(255,223,161))   'dialog color
        Call ReMix                                        'mix and set up new game
        CheckDone

     Case %WM_DESTROY                          'at exit
        If hTimer Then KillTimer hDlg, hTimer  'Kill the timer
        DeleteObject hBrushDialog              'Delete the brush
        PostQuitMessage 0

     Case %WM_COMMAND
        Select Case CbCtl
           Case %ID_MIX                             'Start a new game..
              rStr = "Break current game and start a new?"
              If hTimer = 0 Or MessageBox(hDlg, ByVal StrPtr(rStr), _
                                         Format$(n) & " Tiles" & Chr$(0), _
                                         %MB_YESNO Or %MB_ICONQUESTION) = %IDYES Then
                 Call NewGame
              End If

           Case %IDCANCEL
              If hTimer = 0 Or EndProcedure Then Dialog End CbHndl      'exit program

           Case 100 To 100 + n 'The bricks - label controls
              Local bc As Long, tmp As Long : bc = CbCtl - 100

              'check to see if the brick has an empty space next to it
              If (bPos(bc).z  + 1 = EmptyPos And (bPos(bc).z + 1) Mod brc <> 0) Or _
                 (bPos(bc).z  - 1 = EmptyPos And (EmptyPos + 1) Mod brc <> 0) Or _
                  bPos(bc).z  + brc = EmptyPos Or bPos(bc).z  - brc = EmptyPos Then

                 If hTimer = 0 Then
                    hTimer = SetTimer(hDlg, %ID_TIMER, 1000, ByVal %NULL) '1000 mSec = 1 sec
                 End If

                 'okay, so move the Brick into the empty space
                 Control Set Loc hDlg, CbCtl, bPos(EmptyPos).x, bPos(EmptyPos).y

                 If EmptyPos = bc  Then       'Was it a "good" position?
                    bPos(bc).c = %Red : Beep  'Could play something else here too..
                 Else
                    bPos(bc).c = %Black
                 End If

                 tmp = EmptyPos               'store new empty space
                 EmptyPos = bPos(bc).z
                 bPos(bc).z = tmp
                 bPos(bc).p = bc

                'following triggers %WM_CTLCOLORSTATIC (redraws brick)
                InvalidateRect CbLParam, ByVal %NULL, 0 : UpdateWindow CbLParam

                bc = CheckDone
                If bc >= n Then           'All is done, build and show a message
                   rStr = "Done in "
                   rStr = rStr & Format$(gTime \ 60, "0") & " min. " _
                               & Format$(gTime Mod 60, "0") & " sec."
                   rStr = rStr & Chr$(10, 10)
                   rStr = rStr & "Play again?" & Chr$(0)

                   If MessageBox(hDlg, ByVal StrPtr(rStr), Format$(n) & " Tiles" & Chr$(0), _
                                       %MB_YESNO Or %MB_ICONWARNING) = %IDYES Then
                      Call NewGame
                  Else
                     Dialog End CbHndl     'exit game
                  End If
               End If
            End If
        End Select

     Case %WM_CTLCOLORSTATIC                  'color tiles and set their textcolor
        If GetDlgCtrlID(CbLParam) > 99 Then
           SetTextColor CbWParam, bPos(GetDlgCtrlID(CbLParam) - 100).c
           Function = GetSysColorBrush(%COLOR_INFOBK)
        End If

     Case %WM_CTLCOLORDLG : Function = hBrushDialog 'paint the dialog's background

     Case %WM_SYSCOMMAND
        If CbWParam = %SC_CLOSE Then
           If EndProcedure = 0 Then
              Function = 1: Exit Function
           End If
        End If

     Case %WM_TIMER
        Incr gTime
        Control Set Text hDlg, %ID_LABEL3, _
                    Format$(gTime \ 60, "0") & " min. " & _
                    Format$(gTime Mod 60, "0") & " sec."

  End Select
End Function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
' NewGame generates a new game.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Sub NewGame    'We get smother action with redraw turned off
  If hTimer Then KillTimer hDlg, hTimer  'Kill eventual timer
  hTimer = 0

  Call SendMessage(hDlg, %WM_SETREDRAW, %FALSE, 0)
  Call ReMix   'Set new random order
  Call SendMessage(hDlg, %WM_SETREDRAW, %TRUE, 0)
  InvalidateRect hDlg, ByVal %NULL, 0 : UpdateWindow hDlg
  Call CheckDone
End Sub

'
' ReMix generates random numbers and moves the bricks to their new positions.
' This is only (?) sure way to mix the bricks - mimics real life mix by moving
' bricks around randomly, using real empty space as target space. Cannot simply
' use a bunch of random numbers, since some combinations are impossible to solve..
'
Sub ReMix
  Local aa As Long, bb As Long, tmp As Long, ePos As Long

  For aa = 0 To n - 1       ' Make sure all numbers are unique
     bPos(aa).z = aa
     bPos(aa).c = %Black
  Next
  ePos = n

  Randomize                 ' Seed the random number generator
  For aa = 1 To n * 100     'experiment if you like - think this is enough to mix them properly..
     Do
        bb = Rnd(1, n) - 1 ' Generate number, then see if it can be moved

        If (bPos(bb).z  + 1 = ePos And (bPos(bb).z + 1) Mod brc <> 0) Or _  'test right of pos 1-3
           (bPos(bb).z  - 1 = ePos And (ePos + 1) Mod brc <> 0) Or _        'test left of position 2-4
            bPos(bb).z  + brc = ePos Or bPos(bb).z - brc = ePos Then        'test above/under

            tmp = bPos(bb).z
            bPos(bb).z = ePos
            ePos = tmp
            Exit Do
        End If
     Loop
     If aa > n * 50 And ePos = n Then Exit For 'enough many draws and empty pos in down/right corner - use it!
  Next
  EmptyPos = ePos

  For aa = 0 To n - 1 ' Re-position the bricks
     If aa = bPos(aa).z Then bPos(aa).c = %Red
     Control Set Loc hDlg, 100 + aa, bPos(bPos(aa).z).x, bPos(bPos(aa).z).y
     bPos(aa).p = aa
  Next aa

  gTime = 0                                     ' Reset the timing of the game
  Control Set Text hDlg, %ID_LABEL3, "0 min. 0 sec."
  SetFocus hDlg
End Sub

'
' CheckDone loops through all bricks to see which ones are
' placed correctly and then shows the result in %ID_LABEL1
'
Function CheckDone As Long
  Local aa As Long, isDone As Long

  For aa = 0 To n - 1
     If aa = bPos(aa).z Then
        bPos(aa).c = %Red    'in correct place, use red text color
        Incr isDone          'count correct ones
     Else
        bPos(aa).c = %Black  'not in correct place, use black text
     End If
  Next aa

  If isDone = 1 Then
     Control Set Text hDlg, %ID_LABEL1, Format$(isDone) & " tile done"
  Else
     Control Set Text hDlg, %ID_LABEL1, Format$(isDone) & " tiles done"
  End If
  Function = isDone          'return number of tiles in correct place
End Function

'
' message before end..
'
Function EndProcedure As Long
  If MessageBox(hDlg, "Exit game?", Format$(n) & " Tiles" & Chr$(0), _
                %MB_YESNO Or %MB_ICONQUESTION) = %IDYES Then
     Function = 1 : Exit Function
  End If
End Function
